"""
Functions to connect to and process data from SolarForecastArbiter API
"""
import datetime as dt
import json
import logging


import requests
from urllib3 import Retry
import numpy as np
import pandas as pd


from solarforecastarbiter import datamodel
from solarforecastarbiter.io.utils import (
    json_payload_to_observation_df,
    json_payload_to_forecast_series,
    observation_df_to_json_payload,
    forecast_object_to_json,
    adjust_timeseries_for_interval_label,
    serialize_timeseries,
    HiddenToken, ensure_timestamps,
    load_report_values)


BASE_URL = 'https://api.solarforecastarbiter.org'
logger = logging.getLogger(__name__)


def request_cli_access_token(user, password, **kwargs):
    """Request an API access token from Auth0.

    Parameters
    ----------
    user : str
        Username
    password : str
        Password
    kwargs
        Passed to requests.post. Useful for handling SSL certificates,
        navigating proxies, or other network complications. See requests
        documentation for details.

    Returns
    -------
    access_token : str
    """
    req = requests.post(
        'https://solarforecastarbiter.auth0.com/oauth/token',
        data={'grant_type': 'password', 'username': user,
              'audience': BASE_URL,
              'password': password,
              'client_id': 'c16EJo48lbTCQEhqSztGGlmxxxmZ4zX7'},
        **kwargs)
    req.raise_for_status()
    return req.json()['access_token']


class APISession(requests.Session):
    """
    Subclass of requests.Session to handle requets to the SolarForecastArbiter
    API. The Session provides connection pooling, automatic retries for certain
    types of requets, default timeouts, and a default base url. Responses are
    converted into the appropriate class from datamodel.py or a pandas object.

    Parameters
    ----------
    access_token : string or HiddenToken
        The base64 encoded Bearer token to authenticate with the API
    default_timeout : float or tuple, optional
        A default timeout to add to all requests. If a tuple, the first element
        is the connection timeout and the second is the read timeout.
        Default is 10 seconds for connection and 60 seconds to read from the
        server.
    base_url : string
        URL to use as the base for endpoints to APISession

    Notes
    -----
    To pass the API calls through a proxy server, set either the HTTP_PROXY or
    HTTPS_PROXY environment variable. If necessary, you can also specify a SSL
    certificate using the REQUESTS_CA_BUNDLE environment variable. For example,
    on a Linux machine:

    >>> export HTTPS_PROXY=https://some_corporate_proxy.com:8080
    >>> export REQUESTS_CA_BUNDLE=/path/to/certificates/cert.crt
    >>> python script_that_calls_api.py

    For more information, see the "Advanced Usage" documentation for the
    requests package: https://requests.readthedocs.io/en/master/user/advanced/

    """

    def __init__(self, access_token, default_timeout=(10, 60),
                 base_url=None):
        super().__init__()
        if isinstance(access_token, HiddenToken):
            access_token = access_token.token
        self.headers = {'Authorization': f'Bearer {access_token}',
                        'Accept': 'application/json',
                        'Accept-Encoding': 'gzip,deflate'}
        self.default_timeout = default_timeout
        self.base_url = base_url or BASE_URL
        # set requests to automatically retry
        retries = Retry(total=10, connect=3, read=3, status=3,
                        status_forcelist=[408, 423, 444, 500, 501, 502, 503,
                                          504, 507, 508, 511, 599],
                        backoff_factor=0.5,
                        raise_on_status=False,
                        remove_headers_on_redirect=[])
        adapter = requests.adapters.HTTPAdapter(max_retries=retries)
        self.mount(self.base_url, adapter)

    def request(self, method, url, *args, **kwargs):
        """
        Modify the default Session.request to add in the default timeout
        and make requests relative to the base_url. Users will likely
        use the standard get and post methods instead of calling this directly.

        Raises
        ------
        requests.exceptions.HTTPError
            When an error is encountered in when making the request to the API
        """
        if url.startswith('/'):
            url = f'{self.base_url}{url}'
        else:
            url = f'{self.base_url}/{url}'
        # set a defautl timeout so we never hang indefinitely
        if 'timeout' not in kwargs:
            kwargs['timeout'] = self.default_timeout

        result = super().request(method, url, *args, **kwargs)
        if result.status_code >= 400:
            raise requests.exceptions.HTTPError(
                f'{result.status_code} API Request Error: {result.reason} for '
                f'url: {result.url} and text: {result.text}',
                response=result)

        return result

    def _process_site_dict(self, site_dict):
        if (
                site_dict.get('modeling_parameters', {}).get(
                    'tracking_type', '') in ('fixed', 'single_axis')
        ):
            return datamodel.SolarPowerPlant.from_dict(site_dict)
        else:
            return datamodel.Site.from_dict(site_dict)

    def get_site(self, site_id):
        """
        Retrieve site metadata for site_id from the API and process
        into the proper model.

        Parameters
        ----------
        site_id : string
            UUID of the site to retrieve metadata for

        Returns
        -------
        datamodel.Site or datamodel.SolarPowerPlant
           Dataclass with all the metadata for the site depending on if
           the Site is a power plant with modeling parameters or not.
        """
        req = self.get(f'/sites/{site_id}')
        site_dict = req.json()
        return self._process_site_dict(site_dict)

    def list_sites(self):
        """
        List all the sites available to a user.

        Returns
        -------
        list of datamodel.Sites and datamodel.SolarPowerPlants
        """
        req = self.get('/sites/')
        return [self._process_site_dict(site_dict)
                for site_dict in req.json()]

    def list_sites_in_zone(self, zone):
        """
        List all the sites available to a user in the given climate zone.

        Parameters
        ----------
        zone : str

        Returns
        -------
        list of datamodel.Sites and datamodel.SolarPowerPlants
        """
        req = self.get(f'/sites/in/{zone}')
        return [self._process_site_dict(site_dict)
                for site_dict in req.json()]

    def search_climatezones(self, latitude, longitude):
        """
        Find all climate zones that the location is in.

        Parameters
        ----------
        latitude : float, degrees North
        longitude : float, degrees East of the Prime Meridian

        Returns
        -------
        list
            A list of the climate zones the location is in
        """
        req = self.get('/climatezones/search',
                       params={'latitude': latitude,
                               'longitude': longitude})
        return [r['name'] for r in req.json()]

    def create_site(self, site):
        """
        Create a new site in the API with the given Site model

        Parameters
        ----------
        site : datamodel.Site or datamodel.SolarPowerPlant
            Site to create in the API

        Returns
        -------
        datamodel.Site or datamodel.SolarPowerPlant
            With the appropriate parameters such as site_id set by the API
        """
        site_dict = site.to_dict()
        for k in ('site_id', 'provider', 'climate_zones'):
            site_dict.pop(k, None)
        site_json = json.dumps(site_dict)
        req = self.post('/sites/', data=site_json,
                        headers={'Content-Type': 'application/json'})
        new_id = req.text
        return self.get_site(new_id)

    def get_observation(self, observation_id):
        """
        Get the metadata from the API for the a given observation_id
        in an Observation object.

        Parameters
        ----------
        observation_id : string
            UUID of the observation to retrieve

        Returns
        -------
        datamodel.Observation
        """
        req = self.get(f'/observations/{observation_id}/metadata')
        obs_dict = req.json()
        site = self.get_site(obs_dict['site_id'])
        obs_dict['site'] = site
        return datamodel.Observation.from_dict(obs_dict)

    def list_observations(self):
        """
        List the observations a user has access to.

        Returns
        -------
        list of datamodel.Observation
        """
        req = self.get('/observations/')
        obs_dicts = req.json()
        if isinstance(obs_dicts, dict):
            obs_dicts = [obs_dicts]
        if len(obs_dicts) == 0:
            return []
        sites = {site.site_id: site for site in self.list_sites()}
        out = []
        for obs_dict in obs_dicts:
            obs_dict['site'] = sites.get(obs_dict['site_id'])
            out.append(datamodel.Observation.from_dict(obs_dict))
        return out

    def create_observation(self, observation):
        """
        Create a new observation in the API with the given Observation model

        Parameters
        ----------
        observation : datamodel.Observation
            Observation to create in the API

        Returns
        -------
        datamodel.Observation
            With the appropriate parameters such as observation_id set by the
            API
        """
        obs_dict = observation.to_dict()
        obs_dict.pop('observation_id')
        obs_dict.pop('provider')
        site = obs_dict.pop('site')
        obs_dict['site_id'] = site['site_id']
        obs_json = json.dumps(obs_dict)
        req = self.post('/observations/', data=obs_json,
                        headers={'Content-Type': 'application/json'})
        new_id = req.text
        return self.get_observation(new_id)

    def _process_fx(self, fx_dict, sites={}):
        if fx_dict['site_id'] is not None:
            if fx_dict['site_id'] in sites:
                fx_dict['site'] = sites[fx_dict['site_id']]
            else:
                fx_dict['site'] = self.get_site(fx_dict['site_id'])
        elif fx_dict['aggregate_id'] is not None:
            fx_dict['aggregate'] = self.get_aggregate(fx_dict['aggregate_id'])

        if fx_dict['variable'] == "event":
            return datamodel.EventForecast.from_dict(fx_dict)
        else:
            return datamodel.Forecast.from_dict(fx_dict)

    def get_forecast(self, forecast_id):
        """
        Get Forecast metadata from the API for the given forecast_id

        Parameters
        ----------
        forecast_id : string
            UUID of the forecast to get metadata for

        Returns
        -------
        datamodel.Forecast
        """
        req = self.get(f'/forecasts/single/{forecast_id}/metadata')
        fx_dict = req.json()
        return self._process_fx(fx_dict)

    def list_forecasts(self):
        """
        List all Forecasts a user has access to.

        Returns
        -------
        list of datamodel.Forecast
        """
        req = self.get('/forecasts/single/')
        fx_dicts = req.json()
        if isinstance(fx_dicts, dict):
            fx_dicts = [fx_dicts]
        if len(fx_dicts) == 0:
            return []
        sites = {site.site_id: site for site in self.list_sites()}
        out = []
        for fx_dict in fx_dicts:
            out.append(self._process_fx(fx_dict, sites=sites))
        return out

    def create_forecast(self, forecast):
        """
        Create a new forecast in the API with the given Forecast model

        Parameters
        ----------
        forecast : datamodel.Forecast
            Forecast to create in the API

        Returns
        -------
        datamodel.Forecast
            With the appropriate parameters such as forecast_id set by the API

        """
        fx_dict = forecast.to_dict()
        fx_dict.pop('forecast_id')
        fx_dict.pop('provider')
        site = fx_dict.pop('site')
        agg = fx_dict.pop('aggregate')
        if site is None and agg is not None:
            fx_dict['aggregate_id'] = agg['aggregate_id']
        else:
            fx_dict['site_id'] = site['site_id']
        fx_json = json.dumps(fx_dict)
        req = self.post('/forecasts/single/', data=fx_json,
                        headers={'Content-Type': 'application/json'})
        new_id = req.text
        return self.get_forecast(new_id)

    def _process_prob_forecast(self, fx_dict, sites={}):
        if fx_dict['site_id'] is not None:
            if fx_dict['site_id'] in sites:
                fx_dict['site'] = sites[fx_dict['site_id']]
            else:
                fx_dict['site'] = self.get_site(fx_dict['site_id'])
        elif fx_dict['aggregate_id'] is not None:
            fx_dict['aggregate'] = self.get_aggregate(fx_dict['aggregate_id'])
        cvs = []
        for constant_value_dict in fx_dict['constant_values']:
            # the API just gets the groups attributes for the
            # single constant value forecasts, so avoid
            # those excess calls
            cv_dict = fx_dict.copy()
            cv_dict.update(constant_value_dict)
            cvs.append(
                datamodel.ProbabilisticForecastConstantValue.from_dict(
                    cv_dict))
        fx_dict['constant_values'] = cvs
        return datamodel.ProbabilisticForecast.from_dict(fx_dict)

    def list_probabilistic_forecasts(self):
        """
        List all ProbabilisticForecasts a user has access to.

        Returns
        -------
        list of datamodel.ProbabilisticForecast
        """
        req = self.get('/forecasts/cdf/')
        fx_dicts = req.json()
        if isinstance(fx_dicts, dict):
            fx_dicts = [fx_dicts]
        if len(fx_dicts) == 0:
            return []
        sites = {site.site_id: site for site in self.list_sites()}
        out = []
        for fx_dict in fx_dicts:
            out.append(self._process_prob_forecast(fx_dict, sites))
        return out

    def get_probabilistic_forecast(self, forecast_id):
        """
        Get ProbabilisticForecast metadata from the API for the given
        forecast_id.

        Parameters
        ----------
        forecast_id : string
            UUID of the forecast to get metadata for

        Returns
        -------
        datamodel.ProbabilisticForecast
        """
        # add /metadata after
        # https://github.com/SolarArbiter/solarforecastarbiter-api/issues/158
        req = self.get(f'/forecasts/cdf/{forecast_id}')
        fx_dict = req.json()
        return self._process_prob_forecast(fx_dict)

    def get_probabilistic_forecast_constant_value(self, forecast_id,
                                                  site=None, aggregate=None):
        """
        Get ProbabilisticForecastConstantValue metadata from the API for
        the given forecast_id.

        Parameters
        ----------
        forecast_id : string
            UUID of the forecast to get metadata for
        site : datamodel.Site or None
            If provided, the object will be attached to the returned
            value (faster). If None, object will be created from site
            metadata obtained from the database (slower).
        aggregate : datamodel.Aggregate or None
            If provided and the forecast is of an aggregate, the object
            will be attached to the return value.

        Returns
        -------
        datamodel.ProbabilisticForecastConstantValue

        Raises
        ------
        ValueError
            If provided site.site_id does not match database record of
            forecast object's linked site_id.
        """
        # add /metadata after
        # https://github.com/SolarArbiter/solarforecastarbiter-api/issues/158
        req = self.get(f'/forecasts/cdf/single/{forecast_id}')
        fx_dict = req.json()

        agg_id = fx_dict['aggregate_id']
        site_id = fx_dict['site_id']
        if site_id is not None:
            if site is None:
                site = self.get_site(site_id)
            elif site.site_id != site_id:
                raise ValueError('Supplied site.site_id does not match site_id'
                                 f'from database. site.site_id: {site.site_id}'
                                 f' database site_id: {site_id}')
            fx_dict['site'] = site
        elif agg_id is not None:
            if aggregate is None:
                aggregate = self.get_aggregate(agg_id)
            elif aggregate.aggregate_id != agg_id:
                raise ValueError(
                    'Supplied aggregate.aggregate_id does not match '
                    'aggregate from database. aggregate.aggregate_id: '
                    f'{aggregate.aggregate_id}'
                    f' database aggregate_id: {agg_id}')
            fx_dict['aggregate'] = aggregate
        return datamodel.ProbabilisticForecastConstantValue.from_dict(fx_dict)

    def create_probabilistic_forecast(self, forecast):
        """
        Create a new forecast in the API with the given
        ProbabilisticForecast model

        Parameters
        ----------
        forecast : datamodel.ProbabilisticForecast
            Probabilistic forecast to create in the API

        Returns
        -------
        datamodel.ProbabilisticForecast
            With the appropriate parameters such as forecast_id set by the API
        """
        fx_dict = forecast.to_dict()
        fx_dict.pop('forecast_id')
        fx_dict.pop('provider')
        site = fx_dict.pop('site')
        agg = fx_dict.pop('aggregate')
        if site is None and agg is not None:
            fx_dict['aggregate_id'] = agg['aggregate_id']
        else:
            fx_dict['site_id'] = site['site_id']

        # fx_dict['constant_values'] is tuple of dict representations of
        # all ProbabilisticForecastConstantValue objects in the
        # ProbabilisticForecast. We need to extract just the numeric
        # values from these dicts and put them into a list for the API.
        constant_values_fxs = fx_dict.pop('constant_values')
        constant_values = [fx['constant_value'] for fx in constant_values_fxs]
        fx_dict['constant_values'] = constant_values
        fx_json = json.dumps(fx_dict)
        req = self.post('/forecasts/cdf/', data=fx_json,
                        headers={'Content-Type': 'application/json'})
        new_id = req.text
        return self.get_probabilistic_forecast(new_id)

    def get_observation_time_range(self, observation_id):
        """
        Get the miniumum and maximum timestamps for observation values.

        Parameters
        ----------
        observation_id : string
            UUID of the observation object.

        Returns
        -------
        tuple of (pandas.Timestamp, pandas.Timestamp)
            The minimum and maximum timestamps for values of the observation.
            Values without an explicit timezone from the API are assumed to be
            UTC.
        """
        req = self.get(f'/observations/{observation_id}/values/timerange')
        data = req.json()
        mint = pd.Timestamp(data['min_timestamp'])
        if mint.tzinfo is None and pd.notna(mint):
            mint = mint.tz_localize('UTC')
        maxt = pd.Timestamp(data['max_timestamp'])
        if maxt.tzinfo is None and pd.notna(maxt):
            maxt = maxt.tz_localize('UTC')
        return mint, maxt

    @ensure_timestamps('start', 'end')
    def get_observation_values_not_flagged(
            self, observation_id, start, end, flag, timezone='UTC'):
        """
        Get the dates where the observation series is NOT flagged with
        the given flag/bitmask.

        Parameters
        ----------
        observation_id : string
            UUID of the observation object.
        start : timelike object
            Start time in interval to retrieve values for
        end : timelike object
            End time of the interval
        flag : int
            Days that are not flagged with this flag are returned. This can
            be a compound flag/bitmask of the flags found in
            :py:mod:`solarforecastarbiter.validation.quality_mapping`,
            in which case days that do not have all flags present
            are returned.
        timezone : str, default "UTC"
            The timezone to localize the data before computing the date

        Returns
        -------
        dates : numpy.array of type datetime64[D]
        """
        req = self.get(f'/observations/{observation_id}/values/unflagged',
                       params={'start': start,
                               'end': end,
                               'timezone': timezone,
                               'flag': flag})
        data = req.json()
        dates = data['dates']
        return np.array([dt.date.fromisoformat(d) for d in dates],
                        dtype='datetime64[D]')

    @ensure_timestamps('start', 'end')
    def get_observation_values(self, observation_id, start, end,
                               interval_label=None):
        """
        Get observation values from start to end for observation_id from the
        API

        Parameters
        ----------
        observation_id : string
            UUID of the observation object.
        start : timelike object
            Start time in interval to retrieve values for
        end : timelike object
            End time of the interval
        interval_label : str or None
            If beginning, ending, adjust the data to return only data that is
            valid between start and end. If None or instant, return any data
            between start and end inclusive of the endpoints.

        Returns
        -------
        pandas.DataFrame
            With a datetime index and (value, quality_flag) columns

        Raises
        ------
        ValueError
            If start or end cannot be converted into a Pandas Timestamp
        """
        req = self.get(f'/observations/{observation_id}/values',
                       params={'start': start, 'end': end})
        out = json_payload_to_observation_df(req.json())
        return adjust_timeseries_for_interval_label(
            out, interval_label, start, end)

    def get_forecast_time_range(self, forecast_id):
        """
        Get the miniumum and maximum timestamps for forecast values.

        Parameters
        ----------
        forecast_id : string
            UUID of the forecast object.

        Returns
        -------
        tuple of (pandas.Timestamp, pandas.Timestamp)
            The minimum and maximum timestamps for values of the forecast.
            Values without an explicit timezone from the API are assumed to be
            UTC.
        """
        req = self.get(f'/forecasts/single/{forecast_id}/values/timerange')
        data = req.json()
        mint = pd.Timestamp(data['min_timestamp'])
        if mint.tzinfo is None and pd.notna(mint):
            mint = mint.tz_localize('UTC')
        maxt = pd.Timestamp(data['max_timestamp'])
        if maxt.tzinfo is None and pd.notna(maxt):
            maxt = maxt.tz_localize('UTC')
        return mint, maxt

    @ensure_timestamps('start', 'end')
    def get_forecast_values(self, forecast_id, start, end,
                            interval_label=None):
        """
        Get forecast values from start to end for forecast_id

        Parameters
        ----------
        forecast_id : string
            UUID of the forecast object
        start : timelike object
            Start of the interval to retrieve values for
        end : timelike object
            End of the interval
        interval_label : str or None
            If beginning, ending, adjust the data to return only data that is
            valid between start and end. If None or instant, return any data
            between start and end inclusive of the endpoints.

        Returns
        -------
        pandas.Series
           With the forecast values and a datetime index

        Raises
        ------
        ValueError
            If start or end cannot be converted into a Pandas Timestamp
        """
        req = self.get(f'/forecasts/single/{forecast_id}/values',
                       params={'start': start, 'end': end})
        out = json_payload_to_forecast_series(req.json())
        return adjust_timeseries_for_interval_label(
            out, interval_label, start, end)

    def get_probabilistic_forecast_constant_value_time_range(
            self, forecast_id):
        """
        Get the miniumum and maximum timestamps for forecast values.

        Parameters
        ----------
        forecast_id : string
            UUID of the constant value forecast object.

        Returns
        -------
        tuple of (pandas.Timestamp, pandas.Timestamp)
            The minimum and maximum timestamps for values of the forecast.
            Values without an explicit timezone from the API are assumed to be
            UTC.
        """
        req = self.get(f'/forecasts/cdf/single/{forecast_id}/values/timerange')
        data = req.json()
        mint = pd.Timestamp(data['min_timestamp'])
        if mint.tzinfo is None and pd.notna(mint):
            mint = mint.tz_localize('UTC')
        maxt = pd.Timestamp(data['max_timestamp'])
        if maxt.tzinfo is None and pd.notna(maxt):
            maxt = maxt.tz_localize('UTC')
        return mint, maxt

    @ensure_timestamps('start', 'end')
    def get_probabilistic_forecast_constant_value_values(
            self, forecast_id, start, end, interval_label=None):
        """
        Get forecast values from start to end for forecast_id

        Parameters
        ----------
        forecast_id : string
            UUID of the forecast object
        constant_value : string
            The variable value or percentile.
        start : timelike object
            Start of the interval to retrieve values for
        end : timelike object
            End of the interval
        interval_label : str or None
            If beginning, ending, adjust the data to return only data that is
            valid between start and end. If None or instant, return any data
            between start and end inclusive of the endpoints.

        Returns
        -------
        pandas.Series
           With the forecast values and a datetime index

        Raises
        ------
        ValueError
            If start or end cannot be converted into a Pandas Timestamp
        """
        req = self.get(f'/forecasts/cdf/single/{forecast_id}/values',
                       params={'start': start, 'end': end})
        out = json_payload_to_forecast_series(req.json())
        return adjust_timeseries_for_interval_label(
            out, interval_label, start, end)

    @ensure_timestamps('start', 'end')
    def get_probabilistic_forecast_values(
            self, forecast_id, start, end, interval_label=None):
        """
        Get all probabilistic forecast values for each from start to end for
        forecast_id

        Parameters
        ----------
        forecast_id : string
            UUID of the forecast object
        start : timelike object
            Start of the interval to retrieve values for
        end : timelike object
            End of the interval
        interval_label : str or None
            If beginning, ending, adjust the data to return only data that is
            valid between start and end. If None or instant, return any data
            between start and end inclusive of the endpoints.

        Returns
        -------
        pandas.DataFrame
           With the forecast values in each column with column names as the
           constant value and a datetime index

        Raises
        ------
        ValueError
            If start or end cannot be converted into a Pandas Timestamp
        """
        df_dict = {}
        prob_fx = self.get_probabilistic_forecast(forecast_id)
        for cv in prob_fx.constant_values:
            df_dict[str(cv.constant_value)] = \
                self.get_probabilistic_forecast_constant_value_values(
                    forecast_id=cv.forecast_id, start=start, end=end,
                    interval_label=interval_label)
        return pd.DataFrame(df_dict)

    def post_observation_values(self, observation_id, observation_df,
                                params=None):
        """
        Upload the given observation values to the appropriate observation_id
        of the API.

        Parameters
        ----------
        observation_id : string
            UUID of the observation to add values for
        observation_df : pandas.DataFrame
            Dataframe with a datetime index and the (required) value and
            quality_flag columns to upload to the API.
        params : dict, list, string, default None
            Parameters passed through POST request. Types are the same as
            Requests <https://2.python-requests.org/en/master/api/#requests.Request>
        """  # NOQA
        json_vals = observation_df_to_json_payload(observation_df)
        self.post(f'/observations/{observation_id}/values',
                  data=json_vals, params=params,
                  headers={'Content-Type': 'application/json'})

    def post_forecast_values(self, forecast_id, forecast_series):
        """
        Upload the given forecast values to the appropriate forecast_id of the
        API

        Parameters
        ----------
        forecast_id : string
            UUID of the forecast to upload values to
        forecast_obj : pandas.Series
            Pandas series with a datetime index that contains the values to
            upload to the API
        """
        json_vals = forecast_object_to_json(forecast_series)
        self.post(f'/forecasts/single/{forecast_id}/values',
                  data=json_vals,
                  headers={'Content-Type': 'application/json'})

    def post_probabilistic_forecast_constant_value_values(self, forecast_id,
                                                          forecast_series):
        """
        Upload the given forecast values to the appropriate forecast_id of the
        API

        Parameters
        ----------
        forecast_id : string
            UUID of the forecast to upload values to
        forecast_obj : pandas.Series
            Pandas series with a datetime index that contains the values to
            upload to the API
        """
        json_vals = forecast_object_to_json(forecast_series)
        self.post(f'/forecasts/cdf/single/{forecast_id}/values',
                  data=json_vals,
                  headers={'Content-Type': 'application/json'})

    def process_report_dict(self, rep_dict):
        """
        Load parameters from rep_dict into a Report object, getting forecasts
        and observations as necessary

        Parameters
        ----------
        rep_dict : dict
            Report dictionary as posted to the the API. See the API schema for
            details

        Returns
        -------
        datamodel.Report
        """
        rep_params = rep_dict['report_parameters'].copy()
        req_dict = {}
        for key in ('report_id', 'status', 'provider'):
            req_dict[key] = rep_dict.get(key, '')
        pairs = []
        for o in rep_params['object_pairs']:
            fx_type = o.get('forecast_type', 'forecast')
            fx_method = self._forecast_get_by_type(fx_type)
            fx = fx_method(o['forecast'])
            norm = o.get('normalization')
            unc = o.get('uncertainty')
            cost = o.get('cost')
            ref_fx = o.get('reference_forecast')
            if ref_fx is not None:
                ref_fx = fx_method(ref_fx)
            if 'observation' in o:
                obs = self.get_observation(o['observation'])
                pair = datamodel.ForecastObservation(
                    fx, obs, normalization=norm, uncertainty=unc,
                    reference_forecast=ref_fx, cost=cost)
            elif 'aggregate' in o:
                agg = self.get_aggregate(o['aggregate'])
                pair = datamodel.ForecastAggregate(
                    fx, agg, normalization=norm, uncertainty=unc,
                    reference_forecast=ref_fx, cost=cost)
            else:
                raise ValueError('must provide observation or aggregate in all'
                                 'object_pairs')
            pairs.append(pair)
        rep_params['object_pairs'] = tuple(pairs)
        req_dict['report_parameters'] = rep_params
        return datamodel.Report.from_dict(req_dict)

    def get_report(self, report_id):
        """
        Get the metadata, and possible raw report if it has processed,
        from the API for the given report_id in a Report object.

        Parameters
        ----------
        report_id : string
            UUID of the report to retrieve

        Returns
        -------
        datamodel.Report
        """
        req = self.get(f'/reports/{report_id}')
        resp = req.json()
        raw = resp.pop('raw_report', None)
        report = self.process_report_dict(resp)
        if raw is not None:
            raw_report = datamodel.RawReport.from_dict(raw)
            processed_fxobs = self.get_raw_report_processed_data(
                report_id, raw_report, resp['values'])
            report = report.replace(raw_report=raw_report.replace(
                processed_forecasts_observations=processed_fxobs))
        return report

    def list_reports(self):
        """
        List the reports a user has access to.  Does not load the raw
        report data, use :py:meth:`~.APISession.get_report`.

        Returns
        -------
        list of datamodel.Report

        """
        req = self.get('/reports')
        rep_dicts = req.json()
        if isinstance(rep_dicts, dict):
            rep_dicts = [rep_dicts]
        if len(rep_dicts) == 0:
            return []
        out = []
        for rep_dict in rep_dicts:
            out.append(self.process_report_dict(rep_dict))
        return out

    def create_report(self, report):
        """
        Post the report request to the API. A completed report should post
        the raw_report with :py:meth:`~.APISession.post_raw_report`.

        Parameters
        ----------
        report : datamodel.Report

        Returns
        -------
        datamodel.Report
           As returned by the API

        """
        report_params = report.report_parameters.to_dict()
        fxobs = report_params.pop('object_pairs')
        object_pairs = []
        for _fo in fxobs:
            d = {'forecast': _fo['forecast']['forecast_id']}
            if 'aggregate' in _fo:
                d['aggregate'] = _fo['aggregate']['aggregate_id']
            else:
                d['observation'] = _fo['observation']['observation_id']
            if _fo['reference_forecast'] is not None:
                d['reference_forecast'] = \
                    _fo['reference_forecast']['forecast_id']
            if (_fo['normalization'] is not None and
                    ~np.isnan(_fo['normalization'])):
                d['normalization'] = str(_fo['normalization'])
            if _fo['uncertainty'] is not None:
                d['uncertainty'] = str(_fo['uncertainty'])
            if _fo['cost'] is not None:
                d['cost'] = _fo['cost']
            object_pairs.append(d)
        report_params['object_pairs'] = object_pairs
        params = {'report_parameters': report_params}
        req = self.post('/reports/', json=params,
                        headers={'Content-Type': 'application/json'})
        new_id = req.text
        return self.get_report(new_id)

    def post_raw_report_processed_data(self, report_id, raw_report):
        """
        Post the processed data that was used to make the report to the
        API.

        Parameters
        ----------
        report_id : str
            ID of the report to post values to
        raw_report : datamodel.RawReport
            The raw report object with processed_forecasts_observations

        Returns
        -------
        tuple
            of datamodel.ProcessedForecastObservation with `forecast_values`
            and `observations_values` replaced with report value IDs for later
            retrieval
        """
        posted_fxobs = []
        for fxobs in raw_report.processed_forecasts_observations:
            fx_data = {
                'object_id': fxobs.original.forecast.forecast_id,
                'processed_values': serialize_timeseries(
                    fxobs.forecast_values)}
            fx_post = self.post(
                f'/reports/{report_id}/values',
                json=fx_data, headers={'Content-Type': 'application/json'})
            if isinstance(fxobs.original, datamodel.ForecastObservation):
                obj_id = fxobs.original.observation.observation_id
            else:
                obj_id = fxobs.original.aggregate.aggregate_id
            obs_data = {
                'object_id': obj_id,
                'processed_values': serialize_timeseries(
                    fxobs.observation_values)}
            obs_post = self.post(
                f'/reports/{report_id}/values',
                json=obs_data, headers={'Content-Type': 'application/json'})
            if fxobs.original.reference_forecast is not None:
                ref_fx_data = {
                    'object_id': fxobs.original.reference_forecast.forecast_id,
                    'processed_values': serialize_timeseries(
                        fxobs.reference_forecast_values)}
                ref_fx_post = self.post(
                    f'/reports/{report_id}/values',
                    json=ref_fx_data,
                    headers={'Content-Type': 'application/json'})
                processed_ref_fx_id = ref_fx_post.text
            else:
                processed_ref_fx_id = None
            processed_fx_id = fx_post.text
            processed_obs_id = obs_post.text
            new_fxobs = fxobs.replace(
                forecast_values=processed_fx_id,
                observation_values=processed_obs_id,
                reference_forecast_values=processed_ref_fx_id)
            posted_fxobs.append(new_fxobs)
        return tuple(posted_fxobs)

    def get_raw_report_processed_data(self, report_id, raw_report,
                                      values=None):
        """
        Load the processed forecast/observation data into the
        datamodel.ProcessedForecastObservation objects of the raw_report.

        Parameters
        ----------
        report_id : str
            ID of the report that values will be loaded from
        raw_report : datamodel.RawReport
            The raw report with processed_forecasts_observations to
            be replaced
        values : list or None
            The report values dict as returned by the API. If None, fetch
            the values from the API for the given report_id

        Returns
        -------
        tuple
           Of datamodel.ProcessedForecastObservation with values loaded into
           `forecast_values` and `observation_values`
        """
        if values is None:
            val_req = self.get(f'/reports/{report_id}/values')
            values = val_req.json()
        return load_report_values(raw_report, values)

    def post_raw_report(self, report_id, raw_report, status='complete'):
        """
        Update the report with the raw report and metrics

        Parameters
        ----------
        report_id : str
           ID of the report to update
        raw_report : datamodel.RawReport
           The raw report object to add to the report
        status : str, default 'complete'
           The new status of the report
        """
        posted_fxobs = self.post_raw_report_processed_data(
            report_id, raw_report)
        raw_dict = raw_report.replace(
            processed_forecasts_observations=posted_fxobs).to_dict()
        self.post(f'/reports/{report_id}/raw',
                  json=raw_dict,
                  headers={'Content-Type': 'application/json'})
        self.update_report_status(report_id, status)

    def update_report_status(self, report_id, status):
        """
        Update the status of the report

        Parameters
        ----------
        report_id : str
           ID of the report to update
        status : str
           New status of the report
        """
        self.post(f'/reports/{report_id}/status/{status}')

    def get_aggregate(self, aggregate_id):
        """
        Get Aggregate metadata from the API for the given aggregate_id

        Parameters
        ----------
        aggregate_id : string
            UUID of the aggregate to get metadata for

        Returns
        -------
        datamodel.Aggregate
        """
        req = self.get(f'/aggregates/{aggregate_id}/metadata')
        agg_dict = req.json()
        for o in agg_dict['observations']:
            o['observation'] = self.get_observation(o['observation_id'])
        return datamodel.Aggregate.from_dict(agg_dict)

    def list_aggregates(self):
        """
        List all Aggregates a user has access to.

        Returns
        -------
        list of datamodel.Aggregate
        """
        req = self.get('/aggregates/')
        agg_dicts = req.json()
        if isinstance(agg_dicts, dict):
            agg_dicts = [agg_dicts]
        if len(agg_dicts) == 0:
            return []
        observations = {obs.observation_id: obs
                        for obs in self.list_observations()}
        out = []
        for agg_dict in agg_dicts:
            for o in agg_dict['observations']:
                o['observation'] = observations.get(o['observation_id'])
            out.append(datamodel.Aggregate.from_dict(agg_dict))
        return out

    def create_aggregate(self, aggregate):
        """
        Create a new aggregate in the API with the given Aggregate model

        Parameters
        ----------
        aggregate : datamodel.Aggregate
            Aggregate to create in the API

        Returns
        -------
        datamodel.Aggregate
            With the parameters aggregate_id and provider set by the API.
        """
        agg_dict = aggregate.to_dict()
        agg_dict.pop('aggregate_id')
        agg_dict.pop('provider')
        agg_dict.pop('interval_value_type')
        observations = []
        for obs in agg_dict.pop('observations'):
            if obs['effective_from'] is not None:
                observations.append(
                    {'observation_id': obs['observation']['observation_id'],
                     'effective_from': obs['effective_from']})
            if obs['effective_until'] is not None:
                observations.append(
                    {'observation_id': obs['observation']['observation_id'],
                     'effective_until': obs['effective_until']})

        agg_json = json.dumps(agg_dict)
        req = self.post('/aggregates/', data=agg_json,
                        headers={'Content-Type': 'application/json'})
        new_id = req.text
        obs_json = json.dumps({'observations': observations})
        self.post(f'/aggregates/{new_id}/metadata', data=obs_json,
                  headers={'Content-Type': 'application/json'})
        return self.get_aggregate(new_id)

    @ensure_timestamps('start', 'end')
    def get_aggregate_values(self, aggregate_id, start, end,
                             interval_label=None):
        """
        Get aggregate values from start to end for aggregate_id from the
        API

        Parameters
        ----------
        aggregate_id : string
            UUID of the aggregate object.
        start : timelike object
            Start time in interval to retrieve values for
        end : timelike object
            End time of the interval
        interval_label : str or None
            If beginning or ending, return only data that is
            valid between start and end. If None, return any data
            between start and end inclusive of the endpoints.

        Returns
        -------
        pandas.DataFrame
            With a datetime index and (value, quality_flag) columns

        Raises
        ------
        ValueError
            If start or end cannot be converted into a Pandas Timestamp
        """
        req = self.get(f'/aggregates/{aggregate_id}/values',
                       params={'start': start, 'end': end})
        out = json_payload_to_observation_df(req.json())
        return adjust_timeseries_for_interval_label(
            out, interval_label, start, end)

    @ensure_timestamps('start', 'end')
    def get_values(self, obj, start, end, interval_label=None):
        """
        Get time series values from start to end for object from the API

        Parameters
        ----------
        obj : datamodel.Observation, datamodel.Aggregate, datamodel.Forecast, datamodel.ProbabilisticForecastConstantValues
            Data model object for which to obtain time series data.
        start : timelike object
            Start time in interval to retrieve values for
        end : timelike object
            End time of the interval
        interval_label : str or None
            If beginning or ending, return only data that is
            valid between start and end. If None, return any data
            between start and end inclusive of the endpoints.

        Returns
        -------
        pandas.Series or pandas.DataFrame
            With a datetime index. If DataFrame, (value, quality_flag)
            columns

        Raises
        ------
        ValueError
            If start or end cannot be converted into a Pandas Timestamp
        """  # noqa: E501
        # order avoids possible issues with inheritance
        if isinstance(obj, datamodel.ProbabilisticForecastConstantValue):
            f = self.get_probabilistic_forecast_constant_value_values
            obj_id = obj.forecast_id
        elif isinstance(obj, datamodel.ProbabilisticForecast):
            f = self.get_probabilistic_forecast_values
            obj_id = obj.forecast_id
        elif isinstance(obj, datamodel.Forecast):
            f = self.get_forecast_values
            obj_id = obj.forecast_id
        elif isinstance(obj, datamodel.Aggregate):
            f = self.get_aggregate_values
            obj_id = obj.aggregate_id
        elif isinstance(obj, datamodel.Observation):
            f = self.get_observation_values
            obj_id = obj.observation_id
        return f(obj_id, start, end, interval_label=interval_label)

    def get_user_info(self):
        """
        Get information about the current user from the API

        Returns
        -------
        dict
        """
        req = self.get('/users/current')
        return req.json()

    def _forecast_get_by_type(self, forecast_type):
        """Returns the appropriate function for requesting forecast metadata
        based on `forecast_type`.

        Parameters
        ----------
        forecast_type: str

        Returns
        -------
        function
        """
        if forecast_type == 'forecast':
            return self.get_forecast
        elif forecast_type == 'event_forecast':
            return self.get_forecast
        elif forecast_type == 'probabilistic_forecast':
            return self.get_probabilistic_forecast
        elif forecast_type == 'probabilistic_forecast_constant_value':
            return self.get_probabilistic_forecast_constant_value
        else:
            raise ValueError('Invalid forecast type.')



import asyncio
from functools import partial
import json
import logging
from pathlib import Path
import signal
import sys


import click
import pandas as pd
import requests
import sentry_sdk


from solarforecastarbiter import __version__
from solarforecastarbiter.io import nwp
from solarforecastarbiter.io.api import request_cli_access_token, APISession
from solarforecastarbiter.io.fetch import start_cluster
from solarforecastarbiter.io.reference_observations import reference_data
from solarforecastarbiter.io.utils import mock_raw_report_endpoints
import solarforecastarbiter.reference_forecasts.main as reference_forecasts
from solarforecastarbiter.validation import tasks as validation_tasks
import solarforecastarbiter.reports.main as reports
from solarforecastarbiter.reports import template


CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])
logger = logging.getLogger(__name__)
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s',
                    level=logging.WARNING)
sentry_sdk.init(send_default_pii=False,
                release=f'solarforecastarbiter-core@{__version__}')
midnight = pd.Timestamp.utcnow().floor('1d')


def handle_exception(exc_type, exc_value, exc_traceback):  # pragma: no cover
    if issubclass(exc_type, KeyboardInterrupt):
        sys.__excepthook__(exc_type, exc_value, exc_traceback)
        return
    logging.error("Uncaught exception",
                  exc_info=(exc_type, exc_value, exc_traceback))


sys.excepthook = handle_exception


def cli_access_token(user, password):
    try:
        token = request_cli_access_token(user, password)
    except requests.HTTPError as e:
        click.echo(click.style(
            e.response.json()['error_description'], fg='red'))
        sys.exit(1)
    else:
        return token


def set_log_level(verbose):
    if verbose == 1:
        loglevel = 'INFO'
    elif verbose > 1:
        loglevel = 'DEBUG'
    else:
        loglevel = 'WARNING'
    logging.getLogger().setLevel(loglevel)


class UTCTimestamp(click.ParamType):
    """Convert a timestamp string to a Pandas Timestamp localized to UTC"""
    name = 'UTCTimestamp'

    def convert(self, value, param, ctx):
        try:
            out = pd.Timestamp(value)
        except ValueError:
            self.fail('%s cannot be converted into a Pandas.Timestamp'
                      % value, param, ctx)
        else:
            if out.tzinfo:
                return out.tz_convert('UTC')
            else:
                return out.tz_localize('UTC')


UTCTIMESTAMP = UTCTimestamp()


def common_options(cmd):
    """Combine common options into one decorator"""
    def wrapper(f):
        decs = [
            click.option('-v', '--verbose', count=True,
                         help='Increase logging verbosity'),
            click.option('-u', '--user', show_envvar=True,
                         help='Username to access API.',
                         envvar='SFA_API_USER',
                         required=True),
            click.option('-p', '--password', show_envvar=True,
                         envvar='SFA_API_PASSWORD',
                         required=True,
                         prompt=True, hide_input=True,
                         help='Password to access API'),
            click.option('--base-url', show_default=True,
                         envvar='SFA_API_BASE_URL',
                         show_envvar=True,
                         default='https://api.solarforecastarbiter.org',
                         help='URL of the SolarForecastArbiter API')
            ]
        for dec in reversed(decs):
            f = dec(f)
        return f
    return wrapper(cmd)


@click.group(context_settings=CONTEXT_SETTINGS)
@click.version_option(__version__)
def cli():
    """
    The SolarForecastArbiter core command line tool
    """
    pass  # pragma: no cover


@cli.command()
@common_options
@click.option('--start', show_default='00:00:00 Yesterday (UTC)',
              type=UTCTIMESTAMP,
              default=lambda: midnight - pd.Timedelta(days=1),
              help='datetime to start validation at')
@click.option('--end', default=lambda: midnight,
              type=UTCTIMESTAMP,
              show_default='00:00:00 Today (UTC)',
              help='datetime to end validation at')
@click.option(
    '--only-missing/--not-only-missing',
    is_flag=True, default=True,
    help='Only apply validation to periods where daily validation is missing')
@click.argument('observation_id', nargs=-1)
def validate(verbose, user, password, start, end, base_url,
             only_missing, observation_id):
    """
    Run the validation tasks for a given set of observations
    """
    set_log_level(verbose)
    token = cli_access_token(user, password)
    if not observation_id:
        logger.info(
            ('Validating observation data from %s to %s for all '
             'observations'), start, end)
        validation_tasks.fetch_and_validate_all_observations(
            token, start, end, only_missing=only_missing,
            base_url=base_url)
    else:
        logger.info(
            ('Validating observation data from %s to %s for '
             'observations:\n\t%s'), start, end, ','.join(observation_id))
        for obsid in observation_id:
            validation_tasks.fetch_and_validate_observation(
                token, obsid, start, end, only_missing=only_missing,
                base_url=base_url)


@cli.group(help=reference_data.CLI_DESCRIPTION)
def referencedata():
    pass  # pragma: no cover


network_opt = click.option(
    '--network', multiple=True,
    help="The networks to act on. Defaults to all.",
    default=reference_data.NETWORK_OPTIONS,
    type=click.Choice(reference_data.NETWORK_OPTIONS))


@referencedata.command(name='init')
@common_options
@network_opt
@click.option(
    '--site-file', type=click.Path(exists=True, resolve_path=True),
    default=reference_data.DEFAULT_SITEFILE,
    help='The file from which to load all of the reference site metadata.')
def referencedata_init(verbose, user, password, base_url, network, site_file):
    """
    Creates sites and observations from a site file.
    """
    set_log_level(verbose)
    token = cli_access_token(user, password)
    # click checks if path exists
    all_sites = pd.read_csv(site_file, comment='#')
    network_filtered_sites = all_sites[all_sites['network'].isin(network)]
    site_dictionaries = reference_data.site_df_to_dicts(network_filtered_sites)
    reference_data.initialize_reference_metadata_objects(
        token, site_dictionaries, base_url)


@referencedata.command(name='update')
@common_options
@network_opt
@click.option('--end', type=UTCTIMESTAMP,
              help='End time (ISO8601) to fetch data for. Default is now')
@click.option('--start', type=UTCTIMESTAMP,
              help=('Start time (ISO8601) to fetch data for. Default is'
                    ' max of last timestamp in API and end - 7 days'))
def referencedata_update(verbose, user, password, base_url, network, start,
                         end):
    """
    Updates reference data for the given period.
    """
    set_log_level(verbose)
    token = cli_access_token(user, password)
    reference_data.update_reference_observations(token, start, end, network,
                                                 base_url)


@cli.command()
@click.option('-v', '--verbose', count=True,
              help='Increase logging verbosity')
@click.option('--chunksize', default=128,
              help='Size of a chunk (in KB) to save at one time')
@click.option('--once', is_flag=True,
              help='Only get one forecast initialization time')
@click.option('--use-tmp', is_flag=True,
              help='Save grib files to /tmp')
@click.option('--netcdf-only', is_flag=True,
              help='Only convert files at save_directory to netcdf')
@click.option('--workers', type=int, default=1,
              help='Number of worker processes')
@click.argument('save_directory', type=click.Path(
    exists=True, writable=True, resolve_path=True, file_okay=False))
@click.argument('model', type=click.Choice([
    'gfs_0p25', 'nam_12km', 'rap', 'hrrr_hourly', 'hrrr_subhourly', 'gefs']))
def fetchnwp(verbose, chunksize, once, use_tmp, netcdf_only, workers,
             save_directory, model):
    """
    Retrieve weather forecasts with variables relevant to solar power
    from the NCEP NOMADS server. The utility function wgrib2 is
    required to convert these forecasts into netCDF format.
    """
    set_log_level(verbose)
    from solarforecastarbiter.io.fetch import nwp
    nwp.check_wgrib2()
    start_cluster(workers, 4)
    basepath = Path(save_directory)
    if netcdf_only:
        path_to_files = basepath
        if (
                not path_to_files.is_dir() or
                len(list(path_to_files.glob('*.grib2'))) == 0
        ):
            logger.error('%s is not a valid directory with grib files',
                         path_to_files)
            sys.exit(1)
        fut = asyncio.ensure_future(nwp.optimize_only(path_to_files, model))
    else:
        logger.info('Fetching NWP forecasts for %s', model)
        fut = asyncio.ensure_future(nwp.run(basepath, model, chunksize,
                                            once, use_tmp))

    loop = asyncio.get_event_loop()

    def bail(ecode):
        fut.cancel()
        sys.exit(ecode)

    loop.add_signal_handler(signal.SIGUSR1, partial(bail, 1))
    for sig in (signal.SIGINT, signal.SIGTERM):
        loop.add_signal_handler(sig, partial(bail, 0))

    loop.run_until_complete(fut)


@cli.command()
@common_options
@click.option('--run-time', type=UTCTIMESTAMP,
              help='Run time for the forecasts',
              show_default='now',
              default=pd.Timestamp.utcnow())
@click.option('--issue-time-buffer', type=str,
              help=('Max time-delta between the run time and next '
                    'initialization time'),
              show_default=True,
              default='10min')
@click.argument('nwp_directory', type=click.Path(
    exists=True, resolve_path=True, file_okay=False),
                required=False)
def referencenwp(verbose, user, password, base_url, run_time,
                 issue_time_buffer, nwp_directory):
    """
    Make the reference NWP forecasts that should be issued around run_time
    """
    set_log_level(verbose)
    token = cli_access_token(user, password)
    issue_buffer = pd.Timedelta(issue_time_buffer)
    nwp.set_base_path(nwp_directory)
    reference_forecasts.make_latest_nwp_forecasts(
        token, run_time, issue_buffer, base_url)


@cli.command()
@common_options
@click.option(
    '--format', default='detect',
    help=('Format of output file. "detect" tries to infer from '
          'the file extension of OUTPUT-FILE'),
    type=click.Choice(['detect', 'pdf', 'html'], case_sensitive=False)
)
@click.option(
    '--serialization-roundtrip', is_flag=True,
    help='Run the raw report through a mock API with serialization'
)
@click.option(
    '--orca-server-url', help=(
        'URL to the plotly orca server to generate PDF images if '
        'orca is not installed locally')
)
@click.argument(
    'report-file', type=click.Path(exists=True, resolve_path=True))
@click.argument(
    'output-file', type=click.Path(resolve_path=True))
def report(verbose, user, password, base_url, report_file, output_file,
           format, serialization_roundtrip, orca_server_url):
    """
    Make a report. See API documentation's POST reports section for
    REPORT_FILE requirements.
    """
    set_log_level(verbose)
    token = cli_access_token(user, password)
    with open(report_file) as f:
        metadata = json.load(f)
    session = APISession(token, base_url=base_url)
    report = session.process_report_dict(metadata)
    if orca_server_url is not None:
        import plotly.io as pio
        pio.orca.config.server_url = orca_server_url
    if serialization_roundtrip:
        with mock_raw_report_endpoints(base_url):
            session.create_report(report)
            reports.compute_report(token, 'no_id', base_url)
            full_report = session.get_report('no_id')
    else:
        data = reports.get_data_for_report(session, report)
        raw_report = reports.create_raw_report_from_data(report, data)
        full_report = report.replace(raw_report=raw_report, status='complete')
    # assumed dashboard url based on api url
    dash_url = base_url.replace('api', 'dashboard')
    if (
            (format == 'detect' and output_file.endswith('.html'))
            or format == 'html'
    ):
        html_report = template.render_html(
            full_report, dash_url,
            with_timeseries=True, body_only=False)
        with open(output_file, 'w') as f:
            f.write(html_report)
    elif (
            (format == 'detect' and output_file.endswith('.pdf'))
            or format == 'pdf'
    ):
        pdf_report = template.render_pdf(full_report, dash_url)
        with open(output_file, 'wb') as f:
            f.write(pdf_report)
    else:
        raise ValueError("Unable to detect format")


@cli.command(context_settings=dict(ignore_unknown_options=True))
@click.argument('pytest_args', nargs=-1, type=click.UNPROCESSED)
def test(pytest_args):  # pragma: no cover
    """Test this installation of solarforecastarbiter"""
    import pytest
    ret_code = pytest.main(
        ['--pyargs', 'solarforecastarbiter'] + list(pytest_args))
    sys.exit(ret_code)


if __name__ == "__main__":  # pragma: no cover
    cli()



import pandas as pd
from matplotlib.colors import rgb2hex


from solarforecastarbiter.datamodel import ALLOWED_VARIABLES, COMMON_NAMES


def format_variable_name(variable):
    """Make a human readable name, with units,  for the variable"""
    return f'{COMMON_NAMES[variable]} ({ALLOWED_VARIABLES[variable]})'


def align_index(df, interval_length, limit=None):
    """
    Align the index to the specified interval_length inserting NaNs
    as appropriate.

    Parameters
    ----------
    df : pandas.DataFrame or pandas.Series
        Pandas object with a datetime index that will be reindexed
    interval_length : pd.Timedelta
        Interval to conform input index to
    limit : pd.Timedelta, default None
        Restricts the output index to `end - limit:end`

    Returns
    -------
    pandas.DataFrame or pandas.Series
       Same type as input `df` with an index with frequency `interval_length`
    """
    period_end = df.index[-1]
    if limit is not None:
        period_start = df.index[
            df.index.get_loc(period_end - limit, method='bfill')]
    else:
        period_start = df.index[0]
    # align the data on the index it should have according to the metadata
    nindex = pd.date_range(start=period_start, end=period_end,
                           freq=interval_length,
                           name='timestamp')
    df = df.reindex(nindex, axis=0)
    return df


def line_or_step(interval_label):
    """
    For a given interval_label, determine the plot_method of the data,
    any kwargs for that plot method, and kwargs for adding a hovertool
    for the data.
    """
    if 'instant' in interval_label:
        plot_method = 'line'
        plot_kwargs = dict()
        hover_kwargs = dict(line_policy='nearest',
                            attachment='horizontal')
    elif interval_label == 'beginning':
        plot_method = 'step'
        plot_kwargs = dict(mode='after')
        hover_kwargs = dict(line_policy='prev',
                            attachment='left',
                            add_line=True)
    elif interval_label == 'ending':
        plot_method = 'step'
        plot_kwargs = dict(mode='before')
        hover_kwargs = dict(line_policy='next',
                            attachment='right',
                            add_line=True)
    elif interval_label == 'event':
        plot_method = 'step'
        plot_kwargs = dict(mode='after')
        hover_kwargs = dict(line_policy='prev',
                            attachment='left',
                            add_line=True)
    else:
        raise ValueError(
            'interval_label must be one of "instant", "beginning", '
            '"event", or "ending"')

    return plot_method, plot_kwargs, hover_kwargs


def line_or_step_plotly(interval_label):
    """
    For a given interval_label, forecast_type determine any kwargs for
    the plot.
    """
    if 'instant' in interval_label:
        plot_kwargs = dict()
    elif interval_label == 'beginning':
        plot_kwargs = dict(line_shape='hv')
    elif interval_label == 'ending':
        plot_kwargs = dict(line_shape='vh')
    elif interval_label == 'event':
        plot_kwargs = dict(line_shape='hv', mode='lines+markers')
    else:
        raise ValueError(
            'interval_label must be one of "instant", "beginning", '
            '"event", or "ending"')

    return plot_kwargs


def distribution_fill_color(color_scaler, percentile):
    """Returns a hex code for shading percentiles.
    Parameters
    ----------
    color_scaler: matplotlib.cm.ScalarMappable

    percentile: float

    Returns
    -------
    str
        Hex value of the color to use for shading.
    """
    normalized_value = percentile / 100
    return rgb2hex(color_scaler.to_rgba(normalized_value))


def percentiles_are_symmetric(constant_values):
    """Determines if a list of percentile constant values are symmetric about
    the 50th percentile.

    Parameters
    ----------
    constant_values: list
        List of float constant values
    Returns
    -------
    bool
        True if percentiles are symmetric about 50.
    """
    constant_values = sorted(constant_values)
    lower_bounds = [cv for cv in constant_values if cv < 50]
    upper_bounds = [cv for cv in constant_values if cv > 50][::-1]
    if len(upper_bounds) != len(lower_bounds):
        return False
    for l, u in zip(lower_bounds, upper_bounds):
        if abs(50 - l) != abs(50 - u):
            return False
    return True



import os
import logging
from functools import partial
import pandas as pd

from solarforecastarbiter.io.fetch import eia
from solarforecastarbiter.io.reference_observations import (
    common, default_forecasts)

from requests.exceptions import HTTPError

logger = logging.getLogger('reference_data')


def initialize_site_observations(api, site):
    """Creates an observation at the site.

    Parameters
    ----------
    api : solarforecastarbiter.io.api.APISession
        API Session object, authenticated for the Reference user.
    site : solarforecastarbiter.datamodel.Site
        The site object for which to create the Observations.

    Notes
    -----
    Currently only creates observations for net load [MW]
    (`f"EBA.{eia_site_id}.D.H"`), but EIA contains other variables that may be
    incorporated later (e.g. solar generation:
    `f"EBA.{eia_site_id}.NG.SUN.H"`).

    """

    sfa_var = "net_load"
    logger.info(f'Creating {sfa_var} at {site.name}')
    try:
        common.create_observation(api, site, sfa_var)
    except HTTPError as e:
        logger.error(f'Could not create Observation for "{sfa_var}" '
                     f'at EIA site {site.name}')
        logger.debug(f'Error: {e.response.text}')


def initialize_site_forecasts(api, site):
    """Creates a forecast at the site.

    Parameters
    ----------
    api : solarforecastarbiter.io.api.APISession
        API Session object, authenticated for the Reference user.
    site : solarforecastarbiter.datamodel.Site
        The site object for which to create the Observations.

    """
    common.create_forecasts(
        api, site, ["net_load"],
        default_forecasts.TEMPLATE_NETLOAD_PERSISTENCE_FORECASTS)


def fetch(api, site, start, end, *, eia_api_key):
    """Retrieve observation data for a EIA site between start and end.

    Parameters
    ----------
    api : solarforecastarbiter.io.APISession
        Unused but conforms to common.update_site_observations call
    site : solarforecastarbiter.datamodel.Site
        Site object with the appropriate metadata.
    start : datetime
        The beginning of the period to request data for.
    end : datetime
        The end of the period to request data for.
    eia_api_key : str
        API key for api.eia.gov

    Returns
    -------
    data : pandas.DataFrame
        All of the requested data as a single DataFrame.

    Notes
    -----
    Currently only fetches observations for net load [MW]
    (`f"EBA.{eia_site_id}.D.H"`), but EIA contains other variables that may be
    incorporated later (e.g. solar generation:
    `f"EBA.{eia_site_id}.NG.SUN.H"`).

    """
    try:
        site_extra_params = common.decode_extra_parameters(site)
    except ValueError:
        return pd.DataFrame()
    eia_site_id = site_extra_params['network_api_id']
    series_id = f"EBA.{eia_site_id}.D.H"   # hourly net load (demand)
    obs_df = eia.get_eia_data(
        series_id, eia_api_key,
        start,
        end
    )
    if obs_df.empty:
        logger.warning(f'Data for site {site.name} contained no '
                       f'entries from {start} to {end}.')
        return pd.DataFrame()
    obs_df = obs_df.rename(columns={"value": "net_load"})
    return obs_df


def update_observation_data(api, sites, observations, start, end):
    """Retrieve data from the network, and then format and post it to each
    observation at the site.

    Parameters
    ----------
    api : solarforecastarbiter.io.api.APISession
        An active Reference user session.
    sites: list of solarforecastarbiter.datamodel.Site
        List of all reference sites.
    observations: list of solarforecastarbiter.datamodel.Observation
        List of all reference observations.
    start : datetime
        The beginning of the period to request data for.
    end : datetime
        The end of the period to request data for.

    Raises
    ------
    KeyError
        If EIA_API_KEY environmental variable is not set.

    """

    eia_api_key = os.getenv("EIA_API_KEY")
    if eia_api_key is None:
        raise KeyError('"EIA_API_KEY" environment variable must be '
                       'set to update EIA observation data.')

    eia_sites = common.filter_by_networks(sites, ['EIA'])
    for site in eia_sites:
        common.update_site_observations(
            api, partial(fetch, eia_api_key=eia_api_key),
            site, observations, start, end)



import datetime
from functools import partial

import pandas as pd
from pandas.testing import assert_frame_equal, assert_series_equal
import pytest

from pvlib.location import Location

from solarforecastarbiter import pvmodel


# from pvlib
@pytest.fixture
def golden_mst():
    return Location(39.742476, -105.1786, 'MST', 1830.14)


# from pvlib
@pytest.fixture
def expected_solpos():
    return pd.DataFrame({'elevation': 39.872046,
                         'apparent_zenith': 50.111622,
                         'azimuth': 194.340241,
                         'apparent_elevation': 39.888378},
                        index=['2003-10-17T12:30:30Z'])


# modified from pvlib
def test_calculate_solar_position(expected_solpos, golden_mst):
    times = pd.date_range(datetime.datetime(2003, 10, 17, 12, 30, 30),
                          periods=1, freq='D', tz=golden_mst.tz)
    solar_position = pvmodel.calculate_solar_position(golden_mst.latitude,
                                                      golden_mst.longitude,
                                                      golden_mst.altitude,
                                                      times)
    expected_solpos.index = times
    assert_frame_equal(expected_solpos,
                       solar_position[expected_solpos.columns],
                       check_less_precise=3)


# modified from pvlib
def test_complete_irradiance_components():
    index = pd.DatetimeIndex(
        ['20190101', '20190101', '20190101', '20190629'], tz='UTC')
    ghi = pd.Series([0, 50, 1000, 1000], index=index)
    zenith = pd.Series([120, 85, 10, 10], index=index)
    expected_dni = pd.Series([
        -0.0, 96.71926718388598, 794.2056511357252, 842.3392765276444
    ], index=index)
    expected_dhi = pd.Series([
        0.0, 41.57036043057922, 217.86011727542888, 170.45774980888152
    ], index=index)

    dni, dhi = pvmodel.complete_irradiance_components(ghi, zenith)

    assert_series_equal(expected_dni, dni, check_names=False)
    assert_series_equal(expected_dhi, dhi, check_names=False)


@pytest.fixture
def times():
    return pd.date_range(start='20190101', periods=2, freq='12H',
                         tz='Etc/GMT+7')


@pytest.fixture
def apparent_zenith(times):
    return pd.Series([163.25196176,  62.70113907], index=times)


@pytest.fixture
def azimuth(times):
    return pd.Series([356.76293291, 178.88940279], index=times)


@pytest.fixture
def ghi(times):
    return pd.Series([0., 458.632494], index=times)


@pytest.fixture
def dni(times):
    return pd.Series([0., 1000.], index=times)


@pytest.fixture
def dhi(times):
    return pd.Series([0., 100.], index=times)


@pytest.fixture
def poa_effective(times):
    return pd.Series([0., 1000.], index=times)


@pytest.fixture
def temp_air(times):
    return pd.Series([0., 40.], index=times)


@pytest.fixture
def wind_speed(times):
    return pd.Series([0., 10.], index=times)


def test_clearsky(golden_mst, apparent_zenith):
    latitude, longitude = golden_mst.latitude, golden_mst.longitude
    elevation = golden_mst.altitude
    out = pvmodel.calculate_clearsky(
        latitude, longitude, elevation, apparent_zenith)
    expected = pd.DataFrame(
        [[  0.        ,   0.        ,   0.        ],  # noqa
         [499.27823691, 963.70500214,  57.2923921 ]], # noqa
        columns=['ghi', 'dni', 'dhi'],
        index=apparent_zenith.index)
    assert_frame_equal(expected, out)


def fixed_or_tracking(system_type, expected_fixed, expected_tracking, out):
    if system_type == 'fixed':
        assert_series_equal(expected_fixed, out)
    elif system_type == 'tracking':
        assert_series_equal(expected_tracking, out)
    else:
        raise ValueError('system_type must be fixed or tracking')


def test_aoi_func_factory(modeling_parameters_system_type, apparent_zenith,
                          azimuth):
    modeling_parameters, system_type = modeling_parameters_system_type
    aoi_func = pvmodel.aoi_func_factory(modeling_parameters)
    assert isinstance(aoi_func, partial)
    out = aoi_func(apparent_zenith, azimuth)
    index = apparent_zenith.index
    expected_fixed = (
        30,
        180,
        pd.Series([166.69070004,  32.70998989], index=index)
    )
    expected_tracking = (
        pd.Series([None, 2.15070219], index=index),
        pd.Series([None, 90.], index=index),
        pd.Series([None, 62.68029182], index=index)
    )
    if system_type == 'fixed':
        assert out[0] == expected_fixed[0]
        assert out[1] == expected_fixed[1]
        assert_series_equal(out[2], expected_fixed[2], check_names=False)
    elif system_type == 'tracking':
        for o, e in zip(out, expected_tracking):
            assert_series_equal(o, e, check_names=False)
    else:
        raise Exception


def test_aoi_func_factory_fail():
    with pytest.raises(TypeError):
        pvmodel.aoi_func_factory(None)


@pytest.fixture
def aoi_func_system_type(modeling_parameters_system_type):
    modeling_parameters, system_type = modeling_parameters_system_type
    return pvmodel.aoi_func_factory(modeling_parameters), system_type


def test_calculate_poa_effective(aoi_func_system_type, apparent_zenith,
                                 azimuth, ghi, dni, dhi):
    aoi_func, system_type = aoi_func_system_type
    out = pvmodel.calculate_poa_effective(
        aoi_func, apparent_zenith, azimuth, ghi, dni, dhi)
    index = apparent_zenith.index
    expected_fixed = pd.Series([0., 1161.768016], index=index)
    expected_tracking = pd.Series([0., 1030.2303106183322], index=index)
    fixed_or_tracking(system_type, expected_fixed, expected_tracking, out)


@pytest.mark.parametrize(
    'dc_capacity,temperature_coefficient,dc_loss_factor,'
    'ac_capacity,ac_loss_factor,expected', (
        (10, -0.5, 0, 10, 0, (0., 7.58301313763377)),
        (5, 0.5, 0, 10, 0, (0., 5.)),  # pvwatts_ac max out is pdc0
        (10, 0.000, 20, 10, 20, (0., 6.1789760000000005)),
        (10, 0.000, -20, 10, -20, (0., 12.)),
        (15, -0.5, 0, 10, 0, (0., 10.)),
    ),
    ids=[
        '10:10 DC:AC, neg temp co, no loss',
        '5:10 DC:AC, pos temp co, no loss',
        '10:10 DC:AC, no temp co, losses',
        '10:10 DC:AC, no temp co, enhancing losses',
        '15:10 DC:AC, neg temp co, no loss',
    ])
def test_calculate_power(dc_capacity, temperature_coefficient, dc_loss_factor,
                         ac_capacity, ac_loss_factor, poa_effective, temp_air,
                         wind_speed, expected):
    out = pvmodel.calculate_power(
        dc_capacity, temperature_coefficient, dc_loss_factor,
        ac_capacity, ac_loss_factor, poa_effective, temp_air,
        wind_speed)
    assert_series_equal(pd.Series(expected, index=poa_effective.index), out)


def test_irradiance_to_power(modeling_parameters_system_type, apparent_zenith,
                             azimuth, ghi, dni, dhi, temp_air, wind_speed):
    modeling_parameters, system_type = modeling_parameters_system_type
    out = pvmodel.irradiance_to_power(
        modeling_parameters, apparent_zenith, azimuth, ghi, dni, dhi,
        temp_air=temp_air, wind_speed=wind_speed)
    index = apparent_zenith.index
    expected_fixed = pd.Series([0., 0.003], index=index)
    expected_tracking = pd.Series([0., 0.00293178], index=index)
    fixed_or_tracking(system_type, expected_fixed, expected_tracking, out)



import subprocess


import pytest


from solarforecastarbiter.io import fetch


def badfun():
    raise ValueError


def bad_subprocess():
    subprocess.run(['cat', '/nowaythisworks'], check=True, capture_output=True)


@pytest.fixture(scope='session')
def startcluster():
    fetch.start_cluster(1)


@pytest.mark.asyncio
@pytest.mark.parametrize('bad,err', [
    (badfun, ValueError),
    (bad_subprocess, subprocess.CalledProcessError)
])
async def test_cluster_error(bad, err, startcluster):
    with pytest.raises(err):
        await fetch.run_in_executor(bad)



"""
Inserts metadata and figures into the report template.
"""
import base64
import json
import logging
from pathlib import Path
import re
import subprocess
import tempfile


from bokeh import __version__ as bokeh_version
from jinja2 import Environment, PackageLoader, select_autoescape, ChoiceLoader
from jinja2.runtime import Undefined
from plotly import __version__ as plotly_version


from solarforecastarbiter import datamodel
from solarforecastarbiter.reports.figures import plotly_figures


logger = logging.getLogger(__name__)


def build_metrics_json(report):
    """Creates a dict from the metrics results in the report.

    Parameters
    ----------
    report: :py:class:`solarforecastarbiter.datamodel.Report`

    Returns
    -------
    str
        The json representing the report metrics. The string will be a string
        representing an empty json array if the report does not have a
        computed raw_report.
    """
    if getattr(report, 'raw_report') is not None:
        df = plotly_figures.construct_metrics_dataframe(
            report.raw_report.metrics,
            rename=plotly_figures.abbreviate)
        return df.to_json(orient="records")
    else:
        return "[]"


def _get_render_kwargs(report, dash_url, with_timeseries):
    """Creates a dictionary of key word template arguments for a jinja2
    report template.

    Parameters
    ----------
    report: :py:class:`solarforecastarbiter.datamodel.Report`
    dash_url: str
        URL of the Solar Forecast arbiter dashboard to use when building links.
    with_timeseries: bool
        Whether or not to include timeseries plots. If an error occurs, sets
        `timeseries_script` to an empty string and `timeseries_div` will
        contain a <div> element for warning the user of the failure.

    Returns
    -------
    kwargs: dict
        Dictionary of template variables to unpack as key word arguments when
        rendering.
    """
    kwargs = dict(
        human_categories=datamodel.ALLOWED_CATEGORIES,
        human_metrics=datamodel.ALLOWED_METRICS,
        report=report,
        category_blurbs=datamodel.CATEGORY_BLURBS,
        dash_url=dash_url,
        metrics_json=build_metrics_json(report),
    )
    report_plots = getattr(report.raw_report, 'plots', None)
    # get plotting library versions used when plots were generated.
    # if plot generation failed, fallback to the curent version
    plot_bokeh = getattr(report_plots, 'bokeh_version', None)
    kwargs['bokeh_version'] = plot_bokeh if plot_bokeh else bokeh_version

    plot_plotly = getattr(report_plots, 'plotly_version', None)
    kwargs['plotly_version'] = plot_plotly if plot_plotly else plotly_version
    if with_timeseries:
        try:
            timeseries_specs = plotly_figures.timeseries_plots(report)
        except Exception:
            logger.exception(
                'Failed to make Plotly items for timeseries and scatterplot')
            timeseries_specs = ('{}', '{}', '{}', False)
        kwargs['timeseries_spec'] = timeseries_specs[0]
        kwargs['scatter_spec'] = timeseries_specs[1]
        if timeseries_specs[2] is not None:
            kwargs['timeseries_prob_spec'] = timeseries_specs[2]

        kwargs['includes_distribution'] = timeseries_specs[3]

    return kwargs


def _pretty_json(value):
    if isinstance(value, Undefined):  # pragma: no cover
        return value
    return json.dumps(value, indent=4, separators=(',', ':'))


def _figure_name_filter(value):
    """replace characters that may cause problems for html/javascript ids"""
    if isinstance(value, Undefined):
        return value
    out = (value
           .replace('^', '-')
           .replace(' ', '-')
           .replace('.', 'dot')
           .replace('%', 'percent')
           .replace('<', 'lt')
           .replace('>', 'gt')
           .replace('=', 'eq')
           .replace('(', 'lp')
           .replace(')', 'rp')
           .replace('/', 'fsl')
           .replace('\\', 'bsl')
           )
    out = re.sub('[^\\w-]', 'special', out)
    return out


def get_template_and_kwargs(report, dash_url, with_timeseries, body_only):
    """Returns the jinja2 Template object and a dict of template variables for
    the report. If the report failed to compute, the template and kwargs will
    be for an error page.

    Parameters
    ----------
    report: :py:class:`solarforecastarbiter.datamodel.Report`
    dash_url: str
        URL of the Solar Forecast arbiter dashboard to use when building links.
    with_timeseries: bool
        Whether or not to include timeseries plots.
    body_only: bool
        When True, returns a div for injecting into another template,
        otherwise returns a full html document with the required
        <html> and <head> tags.

    Returns
    -------
    template: jinja2.environment.Template
    kwargs: dict
        Dictionary of template variables to use as keyword arguments to
        template.render().
    """
    env = Environment(
        loader=ChoiceLoader([
            PackageLoader('solarforecastarbiter.reports', 'templates/html'),
            PackageLoader('solarforecastarbiter.reports', 'templates'),
        ]),
        autoescape=select_autoescape(['html', 'xml']),
        lstrip_blocks=True,
        trim_blocks=True
    )
    env.filters['pretty_json'] = _pretty_json
    env.filters['figure_name_filter'] = _figure_name_filter
    kwargs = _get_render_kwargs(report, dash_url, with_timeseries)
    if report.status == 'complete':
        template = env.get_template('body.html')
    elif report.status == 'failed':
        template = env.get_template('failure.html')
    elif report.status == 'pending':
        template = env.get_template('pending.html')
    else:
        raise ValueError(f'Unknown status for report {report.status}')

    if body_only:
        kwargs['base_template'] = env.get_template('empty_base.html')
    else:
        kwargs['base_template'] = env.get_template('base.html')
    return template, kwargs


def render_html(report, dash_url=datamodel.DASH_URL,
                with_timeseries=True, body_only=False):
    """Create full html file.

    The Solar Forecast Arbiter dashboard will likely use its own
    templates for rendering the full html.

    Parameters
    ----------
    report: :py:class:`solarforecastarbiter.datamodel.Report`
    dash_url: str
        URL of the Solar Forecast arbiter dashboard to use when building links.
    with_timeseries: bool
        Whether or not to include timeseries plots.
    body_only: bool
        When True, returns a div for injecting into another template,
        otherwise returns a full html document with the required
        <html> and <head> tags.

    Returns
    -------
    str
        The rendered html report
    """
    template, kwargs = get_template_and_kwargs(
        report, dash_url, with_timeseries, body_only)
    out = template.render(**kwargs)
    return out


def _link_filter(value):
    """convert html href markup to tex href markup"""
    if isinstance(value, Undefined):  # pragma: no cover
        return value
    match = re.search(
        """<a\\s+(?:[^>]*?\\s+)?href=(["'])(.*?)(["'])>(.*?)<\\/a>""",
        value, re.DOTALL)
    if match:
        new = "\\href{" + match.group(2) + "}{" + match.group(4) + "}"
        out = value[:match.start()] + new + value[match.end():]
        return out
    else:
        return value


def _html_to_tex(value):
    if isinstance(value, Undefined):
        return value
    value = (value
             .replace('<p>', '')
             .replace('</p>', '\n')
             .replace('<em>', '\\emph{')
             .replace('</em>', '}')
             .replace('<code>', '\\verb|')
             .replace('</code>', '|')
             .replace('<b>', '\\textbf{')
             .replace('</b>', '}')
             .replace('<ol>', '\\begin{enumerate}')
             .replace('</ol>', '\\end{enumerate}')
             .replace('<li>', '\\item ')
             .replace('</li>', '\n')
             .replace('</a>', '')
             .replace('<=', '$\\leq$')
             .replace("%", "\\%")
             .replace('W/m^2', '$W/m^2$')
             )
    value = re.sub('\\<a.*\\>', '', value)
    return value


def render_pdf(report, dash_url, max_runs=5):
    """
    Create a PDF report using LaTeX.

    Parameters
    ----------
    report: :py:class:`solarforecastarbiter.datamodel.Report`
    dash_url: str
        URL of the Solar Forecast Arbiter dashboard to use when building links.
    max_runs: int, default 5
        Maximum number of times to run pdflatex

    Returns
    -------
    bytes
        The rendered PDF report

    Notes
    -----
    This code was inspired by the latex package available at
    https://github.com/mbr/latex/ under the following license:

    Copyright (c) 2015, Marc Brinkmann
    All rights reserved.

    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice, this
      list of conditions and the following disclaimer.

    * Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.

    * Neither the name of latex nor the names of its
      contributors may be used to endorse or promote products derived from
      this software without specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
    AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
    IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
    DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
    FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
    DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
    SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
    OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
    OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
    """  # NOQA
    env = Environment(
        loader=ChoiceLoader([
            PackageLoader('solarforecastarbiter.reports', 'templates/pdf'),
            PackageLoader('solarforecastarbiter.reports', 'templates'),
        ]),
        autoescape=False,
        lstrip_blocks=True,
        trim_blocks=True,
        block_start_string='\\BLOCK{',
        block_end_string='}',
        variable_start_string='\\VAR{',
        variable_end_string='}',
        comment_start_string='\\#{',
        comment_end_string='}',
        line_statement_prefix='%-',
        line_comment_prefix='%#'
    )
    env.filters['html_to_tex'] = _html_to_tex
    env.filters['link_filter'] = _link_filter
    env.filters['pretty_json'] = _pretty_json
    kwargs = _get_render_kwargs(report, dash_url, False)
    with tempfile.TemporaryDirectory() as _tmpdir:
        tmpdir = Path(_tmpdir)
        logfile, auxfile = _prepare_latex_support_files(tmpdir, env, kwargs)
        _save_figures_to_pdf(tmpdir, report)
        _compile_files_into_pdf(tmpdir, logfile, auxfile, max_runs)
        return (tmpdir / 'out.pdf').read_bytes()


def _prepare_latex_support_files(tmpdir, env, kwargs):
    template = env.get_template('base.tex')
    tex = template.render(**kwargs)
    texfile = tmpdir / 'out.tex'
    texfile.write_text(tex)
    auxfile = tmpdir / 'out.aux'
    logfile = tmpdir / 'out.log'
    return logfile, auxfile


def _save_figures_to_pdf(tmpdir, report):
    figdir = tmpdir / 'figs'
    figdir.mkdir()
    for fig in report.raw_report.plots.figures:
        name = (
            fig.category + '+' + fig.metric + '+' +
            fig.name
        ).replace('^', '-').replace(' ', '+').replace('_', '+').replace(
            '<=', 'lte').replace('%', 'pct').replace('.', '').replace('/', '-')
        name += '.pdf'
        # handle characters that will cause problems for tex
        figpath = figdir / name
        figpath.write_bytes(base64.a85decode(fig.pdf))


def _compile_files_into_pdf(tmpdir, logfile, auxfile, max_runs):
    args = (
        'pdflatex',
        '-interaction=batchmode',
        '-halt-on-error',
        '-no-shell-escape',
        '-file-line-error',
        'out.tex'
    )
    runs_left = max_runs
    prev_aux = 'nothing to see here'
    # run pdflatex until it settles
    while runs_left > 0:
        try:
            subprocess.run(args, check=True, cwd=str(tmpdir.absolute()))
        except subprocess.CalledProcessError:
            try:
                logger.exception(logfile.read_text())
            except Exception:
                logger.exception('Pdflatex failed and so did reading log')
            raise

        aux = auxfile.read_text()
        if aux == prev_aux:
            break
        else:
            prev_aux = aux
            runs_left -= 1
    else:
        raise RuntimeError(
            f'PDF generation unstable after {max_runs} runs')
